using DifferentialEquations
using PyPlot
PyPlot.ioff()
using PyCall
cycler = pyimport("cycler")
using LaTeXStrings
using Printf
using CSV
using QuadGK
using LinearAlgebra
using Roots

Lpnames = [
L"k_{\rm bind}",
L"d_{\rm V}",    
L"k_{\rm diss}",
L"k_{\rm fuse}",
L"k_{\rm uncoat}",
L"d_{\rm endosome}",
L"d_{\rm gRNA}",
L"k_{\rm transl}f_{\rm ORF1}",
L"d_{\rm NSP}",
L"k_{\rm tr_{(-)}}",
L"K_{\rm NSP}",
L"d_{\rm gRNA_{(-)}}",
L"k_{\rm tr_{(+)}}",
L"k_{\rm complex}",
L"k_{\rm complex}n_{\rm N}",
L"k_{\rm transl}f_{\rm N}", 
L"k_{\rm transl}f_{\rm SP}",
L"d_{\rm N}",
L"d_{\rm SP}",
L"K_{\rm V_{rel}}n_{\rm SP}",
L"K_{\rm N}",
L"k_{\rm assemb}",
L"k_{\rm assemb}n_{\rm SP}",
L"d_{\rm N_{^{-}}gRNA}",
L"k_{\rm release}",
L"d_{\rm assembled}"
]

function SARS_CoV2_replication(du,u,p,t)
    V_free, V_bound, V_endosome, gRNA_plus, NSP,
    gRNA_minus, gRNA, N, SP, N_gRNA,
    V_assembled, V_released, V_cumulative = u

    (k_bind, d_V, k_diss, k_fuse, k_uncoat, d_endosome,
     d_gRNA, k_transl_f_ORF1, d_NSP, k_tr_minus, K_NSP,
     d_gRNA_minus, k_tr_plus, k_complex, k_complex_n_N,
     k_transl_f_N, k_transl_f_SP, d_N, d_SP,
     K_V_rel_n_SP, K_V_rel_n_N, k_assembl, k_assembl_n_SP,
     d_N_gRNA, k_release, d_assembled) = p

    #dV_free
    du[1] = -k_bind*V_free - d_V*V_free + k_diss*V_bound

    #dV_bound
    du[2] = k_bind*V_free - (k_fuse + k_diss + d_V)*V_bound

    #dV_endosome
    du[3] = k_fuse*V_bound - (k_uncoat + d_endosome)*V_endosome

    #dgRNA_plus
    du[4] = k_uncoat*V_endosome - d_gRNA*gRNA_plus

    #dNSP
    du[5] = k_transl_f_ORF1*gRNA_plus - d_NSP*NSP

    #dgRNA_minus
    du[6] = k_tr_minus*gRNA_plus*NSP/(K_NSP + NSP) - d_gRNA_minus*gRNA_minus

    f_assemb = SP/(SP + K_V_rel_n_SP)*N_gRNA
    f_complex = N*gRNA/(N + K_V_rel_n_N)

    #dgRNA
    du[7] = k_tr_plus*gRNA_minus*NSP/(K_NSP + NSP) - k_complex*f_complex - d_gRNA*gRNA

    #dN
    du[8] = k_transl_f_N*gRNA - k_complex_n_N*f_complex - d_N*N

    #dSP
    du[9] = k_transl_f_SP*gRNA - k_assembl_n_SP*f_assemb - d_SP*SP

    #dN_gRNA
    du[10] = k_complex*f_complex - k_assembl*f_assemb - d_N_gRNA*N_gRNA

    #dV_assembled
    du[11] = k_assembl*f_assemb - (k_release + d_assembled)*V_assembled

    #dV_released
    du[12] = k_release*V_assembled - d_V*V_released

    #cumulative virion release
    du[13] = V_released
end

# parameters
pt = CSV.File("parameters_lumped.txt", header=false, delim=' ', ignorerepeated=true)
pnames = String.(pt.Column1)
p = Float64.(pt.Column2)
lb = Float64.(pt.Column3)
ub = Float64.(pt.Column4)
@assert all(lb .<= p .<= ub)

println("Obtaining the numerical solution of the model")

# initial condition
u0 = repeat([0.0],13)
V0 = 10.0
u0[1] = V0

# solve the model
t_final = 24.0
tspan = (0.0, t_final)
prob = ODEProblem(SARS_CoV2_replication,u0,tspan,p)
sol = solve(prob,KenCarp5(),reltol=1e-16,abstol=1e-16)


# local sensitivity analysis
using DiffEqSensitivity

println("Performing local sensitivity analysis towards total number of released virions")

# print cumulative progeny release
J,err = quadgk(t -> 8.0*sol(t,idxs=11), tspan[1], tspan[2])
println("J(u,p) = ", J)

function g(u,p,t)
    p[end-1]*u[11]
end

# calculate sensitivities with adjoint method
res = adjoint_sensitivities(sol,KenCarp5(),g,nothing,
                            sensealg=InterpolatingAdjoint(autodiff=false,autojacvec=false),
                            #sensealg=QuadratureAdjoint(abstol=1e-14,reltol=1e-14,
                            #                           autodiff=false,autojacvec=false),
                            abstol=1e-16,reltol=1e-16,
                            iabstol=1e-16,ireltol=1e-16)
s = vec(Float64.(res[2]))

# print sensitivities
function print_sensitivities(filebase, sensitivities, pnames, parameters, J=nothing)
    io = open(filebase * ".txt", "w")
    if !isnothing(J)
        @printf(io,"# J(u,p) = %f\n",J)
    end
    @printf(io,"#% 14s % 15s % 15s\n","parameter","dJ/dp","p*dJ/dp")
    for (pname, s, p) in zip(pnames, sensitivities, parameters)
        @printf(io,"% 15s % 15.5e % 15.5e\n", pname, s, p*s)
    end
    close(io)
end
print_sensitivities("sensitivities_progeny_released", s, pnames, p, J)

function output_sorted_sensitivities(filebase, s, pnames, p, J_tex = L"$J(\mathbf{p})$"; normalized=true)

    # sort sensitivities
    iposs = findall(x -> x>=0, s)
    inegs = findall(x -> x<0, s)

    if !normalized
        p = ones(size(p))
        filebase = filebase * "_unnormed"
    end

    isortedposs = sortperm(s[iposs].*p[iposs], rev=true)
    isortednegs = sortperm(-s[inegs].*p[inegs], rev=true)

    io = open(filebase * "_pos.txt", "w")
    @printf(io,"#% 14s % 15s % 15s\n","parameter","dJ/dp","p*dJ/dp")
    for (i,s_i) in enumerate(s[iposs[isortedposs]])
        @printf(io,"% 15s % 15.5e % 15.5e\n",pnames[iposs[isortedposs]][i],s_i,p[iposs[isortedposs]][i]*s_i)
    end
    close(io)
    io = open(filebase * "_neg.txt", "w")
    @printf(io,"#% 14s % 15s % 15s\n","parameter","dJ/dp","p*dJ/dp")

    for (i,s_i) in enumerate(s[inegs[isortednegs]])
        @printf(io,"% 15s % 15.5e % 15.5e\n",pnames[inegs[isortednegs]][i],s_i,p[inegs[isortednegs]][i]*s_i)
    end
    close(io)

    # draw sensitivities
    PyPlot.rc("text", usetex=false)
    PyPlot.rc("font", size=21)
    logstyleopts = (false, false)

    for logstyle in logstyleopts
        fig, (ax1, ax2) = subplots(1,2,figsize=(23,10), sharex=true)

        if logstyle        
            xshift = -0.02
        else
            xshift = -0.08
        end

        xp = Float64.(range(length(s[inegs[isortednegs]]), 1, step=-1))
        ax1.barh(xp, -s[inegs[isortednegs]].*p[inegs[isortednegs]], color="#36AD4B",
                align="center", log=logstyle, tick_label=pnames[inegs[isortednegs]], zorder=3)
        ax1.tick_params(labeltop=true, top=true, labelbottom=false, bottom=false)
        if normalized
            ax1.set_xlabel(latexstring("Normalized sensitivities (negative effect on $J_tex)"))
            ax1.set_title(L"-p\frac{d\Phi}{dp}", x=xshift-0.01, y=0.997)
        else
            ax1.set_xlabel(latexstring("Sensitivities (negative effect on $J_tex)"))
            ax1.set_title(L"-\frac{d\Phi}{dp}", x=xshift-0.01, y=0.997)
        end
        ax1.grid(true, linewidth=0.01, color="lightgray")

        xp = Float64.(range(length(s[iposs[isortedposs]]), 1, step=-1))
        ax2.barh(xp, s[iposs[isortedposs]].*p[iposs[isortedposs]], color="#EC3E54",
                align="center", log=logstyle, tick_label=pnames[iposs[isortedposs]], zorder=3)
        ax2.tick_params(labeltop=true, top=true, labelbottom=false, bottom=false)
        if normalized
            ax2.set_xlabel(latexstring("Normalized sensitivities (positive effect on $J_tex)"))
            ax2.set_title(L"p\frac{d\Phi}{dp}", x=xshift, y=0.997)
        else
            ax2.set_xlabel(latexstring("Sensitivities (positive effect on $J_tex)"))
            ax2.set_title(L"\frac{d\Phi}{dp}", x=xshift, y=0.997)
        end
        ax2.grid(true, linewidth=0.01, color="lightgray")

        if logstyle
            fname = filebase * ".pdf"
        else
            fname = filebase * "_lin.pdf"
        end
        savefig(fname, bbox_inches="tight")
        plt.close(fig)
    end
end

output_sorted_sensitivities("sensitivities_progeny_released", s, Lpnames, p,
                            L"${\rm \Phi_{progeny}}(\mathbf{p})$")

