using DifferentialEquations
using PyPlot
PyPlot.ioff()
using PyCall
cycler = pyimport("cycler")
using LaTeXStrings
using Printf
using CSV
using QuadGK
using LinearAlgebra
using Roots

Lpnames = [
L"k_{\rm bind}",
L"d_{\rm V}",    
L"k_{\rm diss}",
L"k_{\rm fuse}",
L"k_{\rm uncoat}",
L"d_{\rm endosome}",
L"d_{\rm gRNA}",
L"k_{\rm transl}f_{\rm ORF1}",
L"d_{\rm NSP}",
L"k_{\rm tr_{(-)}}",
L"K_{\rm NSP}",
L"d_{\rm gRNA_{(-)}}",
L"k_{\rm tr_{(+)}}",
L"k_{\rm complex}",
L"k_{\rm complex}n_{\rm N}",
L"k_{\rm transl}f_{\rm N}", 
L"k_{\rm transl}f_{\rm SP}",
L"d_{\rm N}",
L"d_{\rm SP}",
L"K_{\rm V_{rel}}n_{\rm SP}",
L"K_{\rm N}",
L"k_{\rm assemb}",
L"k_{\rm assemb}n_{\rm SP}",
L"d_{\rm N_{^{-}}gRNA}",
L"k_{\rm release}",
L"d_{\rm assembled}"
]

function SARS_CoV2_replication(du,u,p,t)
    V_free, V_bound, V_endosome, gRNA_plus, NSP,
    gRNA_minus, gRNA, N, SP, N_gRNA,
    V_assembled, V_released, V_cumulative = u

    (k_bind, d_V, k_diss, k_fuse, k_uncoat, d_endosome,
     d_gRNA, k_transl_f_ORF1, d_NSP, k_tr_minus, K_NSP,
     d_gRNA_minus, k_tr_plus, k_complex, k_complex_n_N,
     k_transl_f_N, k_transl_f_SP, d_N, d_SP,
     K_V_rel_n_SP, K_V_rel_n_N, k_assembl, k_assembl_n_SP,
     d_N_gRNA, k_release, d_assembled) = p

    #dV_free
    du[1] = -k_bind*V_free - d_V*V_free + k_diss*V_bound

    #dV_bound
    du[2] = k_bind*V_free - (k_fuse + k_diss + d_V)*V_bound

    #dV_endosome
    du[3] = k_fuse*V_bound - (k_uncoat + d_endosome)*V_endosome

    #dgRNA_plus
    du[4] = k_uncoat*V_endosome - d_gRNA*gRNA_plus

    #dNSP
    du[5] = k_transl_f_ORF1*gRNA_plus - d_NSP*NSP

    #dgRNA_minus
    du[6] = k_tr_minus*gRNA_plus*NSP/(K_NSP + NSP) - d_gRNA_minus*gRNA_minus

    f_assemb = SP/(SP + K_V_rel_n_SP)*N_gRNA
    f_complex = N*gRNA/(N + K_V_rel_n_N)

    #dgRNA
    du[7] = k_tr_plus*gRNA_minus*NSP/(K_NSP + NSP) - k_complex*f_complex - d_gRNA*gRNA

    #dN
    du[8] = k_transl_f_N*gRNA - k_complex_n_N*f_complex - d_N*N

    #dSP
    du[9] = k_transl_f_SP*gRNA - k_assembl_n_SP*f_assemb - d_SP*SP

    #dN_gRNA
    du[10] = k_complex*f_complex - k_assembl*f_assemb - d_N_gRNA*N_gRNA

    #dV_assembled
    du[11] = k_assembl*f_assemb - (k_release + d_assembled)*V_assembled

    #dV_released
    du[12] = k_release*V_assembled - d_V*V_released

    #cumulative virion release
    du[13] = V_released
end

# parameters
pt = CSV.File("parameters_lumped.txt", header=false, delim=' ', ignorerepeated=true)
pnames = String.(pt.Column1)
p = Float64.(pt.Column2)
lb = Float64.(pt.Column3)
ub = Float64.(pt.Column4)
@assert all(lb .<= p .<= ub)

println("Obtaining the numerical solution of the model")

# initial condition
u0 = repeat([0.0],13)
V0 = 10.0
u0[1] = V0

# solve the model
t_final = 24.0
tspan = (0.0, t_final)
prob = ODEProblem(SARS_CoV2_replication,u0,tspan,p)
sol = solve(prob,KenCarp5(),reltol=1e-16,abstol=1e-16)


function draw_validation(sol, sol_MOI3, sol_MOI5, filename)
    i2mm = 0.039370077777777776

    PyPlot.ioff()

    PyPlot.rc("text", usetex=false)
    PyPlot.rc("font", size=12)
    #custom_cycler = cycler.cycler(color=["#006BA4", "#FF800E", "#ABABAB", "#595959", "#5F9ED1", "#C85200", "#898989", "#A2C8EC", "#FFBC79", "#CFCFCF"])
    custom_cycler = cycler.cycler(color=["#5170A8", "#F88D00", "#EC3E54", "#67BAB1", "#36AD4B", "#EAD335", "#B76AA1", "#FF8FA4", "#9E745D", "#BAAEAA"])
    #PyPlot.rc("axes", prop_cycler=custom_cycler)

    fig, axes = subplots(1,2,
         figsize=(270.0*i2mm,170.0*i2mm))
         #constrained_layout=true

    ax = axes[1]
    ax.set_prop_cycle(custom_cycler)

    # draw positive- to negative-sense gRNA ratio
    # (Irigoyen et al. PLOS Pathogens. 2016. DOI:10.1371/journal.ppat.1005473)
    t_exp = [1.0, 2.5, 5.0, 8.0]
    ratio_exp_rep1 = [157.3, 14.3, 108.5, 245.4]
    ratio_exp_rep2 = [4282.3, 6.7, 298.8, 396.5]
    tt = 1:0.01:8

    ax.plot(tt, (sol(tt, idxs=4) + sol(tt, idxs=7)) ./ sol(tt, idxs=6),
	    linewidth=2.0,
 	    label=L"($[{\rm gRNA_{(+)}}] + [{\rm gRNA}])/[{\rm gRNA_{(-)}}]$")
    ax.plot(t_exp, ratio_exp_rep1, linestyle="--", marker="o", label="exp. (MOI=10, rep. 1)")
    ax.plot(t_exp, ratio_exp_rep2, linestyle="--", marker="o", label="exp. (MOI=10, rep. 2)")

    ax.set_yscale("log", subs=2:9)
    ax.minorticks_on()
    ax.grid(which="major", linestyle="-", linewidth=0.1, color="lightgray")
    ax.tick_params(width=0.1, which="minor")
    ax.legend()
    ax.set_xlabel("t, hours")
    ax.set_ylabel("ratio")
    ax.set_title("Ratio of positive- to negative-sense vRNA")

    ax = axes[2]
    ax.set_prop_cycle(custom_cycler)

    # draw fold change of released progeny kinetics
    t0 = find_zero(t -> sol_MOI3(t, idxs=12) - 1, 8)
    tt = t0:0.01:48
    ax.plot(tt, sol_MOI3(tt, idxs=12) ./ sol_MOI3(t0, idxs=12),
	    linewidth=2.0,
	    label=L"model ($[{\rm V_{free}}](0) = 10$)")
    t0 = find_zero(t -> sol_MOI5(t, idxs=12) - 1, 8)
    tt = t0:0.01:48
    ax.plot(tt, sol_MOI5(tt, idxs=12) ./ sol_MOI5(t0, idxs=12),
	    linewidth=2.0,
	    label=L"model ($[{\rm V_{free}}](0) = 15$)")

    # (Ogando et al. Journal of General Virology. 2020. DOI:10.1099/jgv.0.001453)
    t_exp1 = [8.0, 10.0, 12.0, 14.0, 24.0, 30.0]
    FC_MOI3 = [1.0, 6.8, 12.2, 48.6, 111.9, 113.8]
    ax.plot(t_exp1, FC_MOI3, linestyle="--", marker="s", label="exp. 1 (MOI=3)")

    # (Hou et al. Cell. 2020. DOI:10.1016/j.cell.2020.05.042)
    t_exp2 = [6.0, 12.0, 18.0, 24.0, 36.0, 48.0]
    FC_MOI5_isolate = [1.0, 90.6, 238.1, 295.1, 554.2, 131.9]
    FC_MOI5_WT = [1.0, 49.6, 42.8, 76.1, 53.0, 6.9]
    FC_MOI5_GFP = [1.0, 66.6, 73.1, 69.3, 40.5, 19.9]
    FC_MOI5_nLuc_GFP = [1.0, 105.0, 155.0, 161.3, 123.4, 65.7]
    ax.plot(t_exp2, FC_MOI5_isolate, linestyle="--", marker="o", label="exp. 2 (MOI=5, isolate)")
    ax.plot(t_exp2, FC_MOI5_WT, linestyle="--", marker="o", label="exp. 2 (MOI=5, WT)")
    ax.plot(t_exp2, FC_MOI5_GFP, linestyle="--", marker="o", label="exp. 2 (MOI=5, GFP)")
    ax.plot(t_exp2, FC_MOI5_nLuc_GFP, linestyle="--", marker="o", label="exp. 2 (MOI=5, nLuc-GFP)")

    ax.set_yscale("log", subs=2:9)
    ax.minorticks_on()
    ax.grid(which="major", linestyle="-", linewidth=0.1, color="lightgray")
    ax.tick_params(width=0.1, which="minor")
    ax.legend()
    ax.set_xlabel("t, hours")
    ax.set_ylabel("fold change")
    ax.set_title("Progeny release kinetics") #"(fold change from the start of release)"

    plt.tight_layout()
    PyPlot.savefig(filename)
    plt.close(fig)
end

tspan = (0.0, 48.0)

u0[1] = 10
prob = ODEProblem(SARS_CoV2_replication,u0,tspan,p)
sol_MOI3 = solve(prob,KenCarp5(),reltol=1e-16,abstol=1e-16)
u0[1] = 15
prob = ODEProblem(SARS_CoV2_replication,u0,tspan,p)
sol_MOI5 = solve(prob,KenCarp5(),reltol=1e-16,abstol=1e-16)

draw_validation(sol, sol_MOI3, sol_MOI5, "validation.pdf")
